import re
# temp: use regex!
def check_pair_validity(pair: str)-> bool:
    assets = pair.split("/")

    if len(assets) != 2:
        return False

    print(assets)
    
    for asset in assets:
        print(asset)
        if not re.match(r"[A-Z]{2,6}", asset) or len(asset) > 6:
            return False

    return True


from typing import Dict, Optional, List, Tuple
import hmac
import hashlib
from operator import itemgetter
import aiohttp
import urllib3
import asyncio
import json
import time
from nexo.exceptions import NexoAPIException, NEXO_API_ERROR_CODES, NexoRequestException
from nexo.helpers import check_pair_validity

class BaseClient:

    API_URL = "https://pro-api.nexo.io/api"
    PUBLIC_API_VERSION = "v1"

    REQUEST_TIMEOUT: float = 10

    def __init__(self, api_key, api_secret):
        self.API_KEY = api_key
        self.API_SECRET = api_secret
        self.timestamp_offset = 0
        
    def _create_api_uri(self, path: str, version: str = PUBLIC_API_VERSION) -> str:
        url = self.API_URL
        return url + '/' + version + '/' + path
    
    def _generate_signature(self, data: Dict) -> str:
        assert self.API_SECRET, "API Secret required for private endpoints"
        ordered_data = self._order_params(data)
        query_string = '&'.join([f"{d[0]}={d[1]}" for d in ordered_data])
        m = hmac.new(self.API_SECRET.encode('utf-8'), query_string.encode('utf-8'), hashlib.sha256).hexdigest()
        return m
    
    def _get_headers(self, url: str, payload: Dict) -> Dict:
        headers = {
            'Accept': 'application/json',
            'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/56.0.2924.87 Safari/537.36',  # noqa
        }
        
        headers['X-API-KEY'] = self.API_KEY
        headers['X-NONCE'] = int(time.time() * 1000 + self.timestamp_offset)
        headers['X-SIGNATURE'] = self._generate_signature(payload)
        print(payload)
        return headers
    
    @staticmethod
    def _order_params(data: Dict) -> List[Tuple[str, str]]:
        """Convert params to list
        :param data:
        :return:
        """
        data = dict(filter(lambda el: el[1] is not None, data.items()))
        params = []
        for key, value in data.items():
            params.append((key, str(value)))
        # sort parameters by key
        params.sort(key=itemgetter(0))
        return params
    
    def _get_request_kwargs(self, method, force_params: bool = False, **kwargs) -> Dict:

        # set default requests timeout
        # kwargs['timeout'] = self.REQUEST_TIMEOUT

        data = kwargs.get('data', None)

        if data and isinstance(data, dict):
            kwargs['data'] = data

        # generate signature
        # kwargs['data']['timestamp'] = int(time.time() * 1000 + self.timestamp_offset)
        # kwargs['data']['signature'] = self._create_hmac_signature(kwargs['data'])

        # sort get and post params to match signature order
        if data:
            # sort post params and remove any arguments with values of None
            kwargs['data'] = self._order_params(kwargs['data'])
            # Remove any arguments with values of None.
            null_args = [i for i, (key, value) in enumerate(kwargs['data']) if value is None]
            for i in reversed(null_args):
                del kwargs['data'][i]

        # if get request assign data array to params value for requests lib
        if data and (method == 'get' or force_params):
            kwargs['params'] = '&'.join('%s=%s' % (data[0], data[1]) for data in kwargs['data'])
            del(kwargs['data'])

        return kwargs
        

class Client(BaseClient):
    def __init__(self, api_key, api_secret):
        super().__init__(api_key, api_secret)
    
    @staticmethod
    def _handle_response(response: urllib3.HTTPResponse):
        """Internal helper for handling API responses from the Binance server.
        Raises the appropriate exceptions when necessary; otherwise, returns the
        response.
        """
        json_response = json.loads(response.data)
        try:
            if "errorCode" in json_response:
                if json_response["errorCode"] in NEXO_API_ERROR_CODES:
                    raise NexoAPIException(json_response["errorCode"], response.data)
                else:
                    raise NexoRequestException(f'Invalid Response: status: {json_response["errorCode"]}, message: {json_response["errorMessage"]}')
            return json_response
        except ValueError:
            raise NexoRequestException('Invalid Response: %s' % json_response)
    
    def _request(self, method, uri: str, **kwargs):

        kwargs = self._get_request_kwargs(method, **kwargs)
        http = urllib3.PoolManager()
        if method == "get" or method == "delete" or method == "head" or method == "options":
            self.response = http.request(method=method.upper(), fields=kwargs, url=uri, headers=self._get_headers(url=f"{uri}", payload=kwargs))
        else:
            self.response = http.request_encode_body(method=method.upper(), fields=kwargs, url=uri, headers=self._get_headers(url=f"{uri}", payload=kwargs), encode_multipart=False)
        
        return self._handle_response(self.response)

    def _request_api(
        self, method, path: str, version=BaseClient.PUBLIC_API_VERSION, **kwargs
    ):
        uri = self._create_api_uri(path, version)
        return self._request(method, uri, **kwargs)
    
    def _get(self, path, version=BaseClient.PUBLIC_API_VERSION, **kwargs):
        return self._request_api('get', path, version, **kwargs)

    def _post(self, path, version=BaseClient.PUBLIC_API_VERSION, **kwargs) -> Dict:
        return self._request_api('post', path, version, **kwargs)

    def _put(self, path, version=BaseClient.PUBLIC_API_VERSION, **kwargs) -> Dict:
        return self._request_api('put', path, version, **kwargs)

    def _delete(self, path, version=BaseClient.PUBLIC_API_VERSION, **kwargs) -> Dict:
        return self._request_api('delete', path, version, **kwargs)

    def get_account_balances(self):
        balances = self._request_api('get', 'accountSummary')
        return balances
    
    def get_pairs(self):
        pairs = self._request_api('get', 'pairs')
        return pairs
    
    def get_price_quote(self, **kwargs):
        if not "pair" in kwargs:
            raise NexoRequestException(f"Bad Request: [pair] is a required parameter")
        
        if not "amount" in kwargs:
            raise NexoRequestException(f"Bad Request: [amount] is a required parameter")
        
        if not "side" in kwargs:
            raise NexoRequestException(f"Bad Request: [side] is a required parameter")
        
        if kwargs["side"] != "buy" and "sell":
            raise NexoRequestException(f"Bad Request: Tried to get price quote with side = {kwargs['side']}, side must be 'buy' or 'sell'")
        if not check_pair_validity(kwargs["pair"]):
            raise NexoRequestException(f"Bad Request: Tried to place a trigger order with pair = {kwargs['pair']}, must be of format [A-Z]{{2,6}}/[A-Z]{{2, 6}}")

        quote = self._request_api('get', 'quote', **kwargs)
        return quote
    
    def get_order_history(self):
        orders = self._request_api('get', 'orders')
        return orders

    def get_order_details(self, **kwargs):
        if not "id" in kwargs:
            raise NexoRequestException(f"Bad Request: [id] is a required parameter")

        order_details = self._request_api('get', 'orderDetails', **kwargs)
        return order_details

    def get_trade_history(self):
        trades = self._request_api('get', 'trades')
        return trades
    
    def get_transaction_info(self, **kwargs):
        if not "transactionId" in kwargs:
            raise NexoRequestException(f"Bad Request: [transactionId] is a required parameter")

        transaction = self._request_api('get', 'transaction', **kwargs)
        return transaction

    def place_order(self, **kwargs):
        if not "pair" in kwargs:
            raise NexoRequestException(f"Bad Request: [pair] is a required parameter")
        if not "side" in kwargs:
            raise NexoRequestException(f"Bad Request: [side] is a required parameter")
        if not "type" in kwargs:
            raise NexoRequestException(f"Bad Request: [type] is a required parameter")
        if not "quantity" in kwargs:
            raise NexoRequestException(f"Bad Request: [quantity] is a required parameter")

        if kwargs["side"] != "buy" and "sell":
            raise NexoRequestException(f"Bad Request: Tried to place an order with side = {kwargs['side']}, side must be 'buy' or 'sell'")
        if kwargs["type"] != "market" and "limit":
            raise NexoRequestException(f"Bad Request: Tried to place an order with type = {kwargs['type']}, side must be 'market' or 'limit'")
        if not check_pair_validity(kwargs["pair"]):
            raise NexoRequestException(f"Bad Request: Tried to place a trigger order with pair = {kwargs['pair']}, must be of format [A-Z]{{2,6}}/[A-Z]{{2, 6}}")

        order_id = self._request_api('post', 'orders', **kwargs)
        return order_id
    
    def place_trigger_order(self, **kwargs):
        if not "pair" in kwargs:
            raise NexoRequestException(f"Bad Request: [pair] is a required parameter")
        if not "side" in kwargs:
            raise NexoRequestException(f"Bad Request: [side] is a required parameter")
        if not "triggerType" in kwargs:
            raise NexoRequestException(f"Bad Request: [triggerType] is a required parameter")
        if not "amount" in kwargs:
            raise NexoRequestException(f"Bad Request: [amount] is a required parameter")
        if not "triggerPrice" in kwargs:
            raise NexoRequestException(f"Bad Request: [triggerPrice] is a required parameter")

        if kwargs["side"] != "buy" and "sell":
            raise NexoRequestException(f"Bad Request: Tried to place a trigger order with side = {kwargs['side']}, side must be 'buy' or 'sell'")
        if kwargs["triggerType"] != "stopLoss" and "takeProfit" and "trailing":
            raise NexoRequestException(f"Bad Request: Tried to place a trigger order with type = {kwargs['triggerType']}, side must be 'market' or 'limit'")
        if not check_pair_validity(kwargs["pair"]):
            raise NexoRequestException(f"Bad Request: Tried to place a trigger order with pair = {kwargs['pair']}, must be of format [A-Z]{{2,6}}/[A-Z]{{2, 6}}")

        order_id = self._request_api('post', 'orders', **kwargs)
        return order_id
    
    def place_advanced_order(self, **kwargs):
        if not "pair" in kwargs:
            raise NexoRequestException(f"Bad Request: [pair] is a required parameter")
        if not "side" in kwargs:
            raise NexoRequestException(f"Bad Request: [side] is a required parameter")
        if not "amount" in kwargs:
            raise NexoRequestException(f"Bad Request: [amount] is a required parameter")
        if not "stopLossPrice" in kwargs:
            raise NexoRequestException(f"Bad Request: [stopLossPrice] is a required parameter")
        if not "takeProfitPrice" in kwargs:
            raise NexoRequestException(f"Bad Request: [takeProfitPrice] is a required parameter")

        if kwargs["side"] != "buy" and "sell":
            raise NexoRequestException(f"Bad Request: Tried to place a trigger order with side = {kwargs['side']}, side must be 'buy' or 'sell'")

        if not check_pair_validity(kwargs["pair"]):
            raise NexoRequestException(f"Bad Request: Tried to place a trigger order with pair = {kwargs['pair']}, must be of format [A-Z]{{2,6}}/[A-Z]{{2, 6}}")

        order_id = self._request_api('post', 'orders', **kwargs)
        return order_id
    
    def place_twap_order(self, **kwargs):
        twap_order = self._request_api('post', 'orders/twap', **kwargs)
        return twap_order


import json

NEXO_API_ERROR_CODES = {
    100: "API-Key is malformed or invalid.",
    101: "Request signature is malformed or invalid.",
    102: "Some request field is malformed or missing.",
    103: "Unauthorized.",
    104: "Websocket method is invalid.",
    105: "Websocket session is already authenticated.",
    106: "Request nonce is malformed or invalid.",
    203: "No results found for specified query.",
    206: "Internal error.",
    300: "The given exchanges are unsupported for said pair.",
    301: "Rate limit exceeded.",
}

class NexoAPIException(Exception):
    def __init__(self, status_code: int, response: str):
        self.code = 0
        try:
            json_res = json.loads(response)
        except ValueError:
            self.message = 'Invalid JSON error message from Nexo: {}'.format(response.text)
        else:
            self.code = json_res['errorCode']
            self.message = json_res['errorMessage']
        self.status_code = status_code
        self.response = response
        self.request = getattr(response, 'request', None)

    def __str__(self):  # pragma: no cover
        return f'APIError(code={self.code}): {self.message}, {NEXO_API_ERROR_CODES[self.code]}'

class NexoRequestException(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return 'NexoRequestException: %s' % self.message

class NotImplementedException(Exception):
    def __init__(self, value):
        message = f'Not implemented: {value}'
        super().__init__(message)



if __name__ == "__main__":
    from dotenv import load_dotenv
    import os
    from client import Client
    load_dotenv()

    nexo_key = os.getenv("NEXO_PUBLIC_KEY")
    nexo_secret = os.getenv("NEXO_SECRET_KEY")

    assert(nexo_key)
    assert(nexo_secret)

    c = Client(nexo_key, nexo_secret)
    balances = c.place_order(pair="BTC/BSD", side="buy", type="market", quantity=10.0, price=30.0)
    print(balances)


#!/usr/bin/env python
from setuptools import setup
import codecs
import os
import re

with open("requirements.txt") as f:
    required = f.read().splitlines()


with codecs.open(
        os.path.join(
            os.path.abspath(os.path.dirname(__file__)),
            'nexo',
            '__init__.py'
        ), 'r', 'latin1') as fp:
    try:
        version = re.findall(r"^__version__ = '([^']+)'\r?$", fp.read(), re.M)[0]
    except IndexError:
        raise RuntimeError('Unable to determine version.')

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='python-nexo',
    version=version,
    packages=['nexo'],
    description='Nexo Pro REST API python implementation',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/guilyx/python-nexo',
    author='Erwin Lejeune',
    license='MIT',
    author_email='erwin.lejeune15@gmail.com',
    install_requires=required,
    keywords='nexo crypto exchange rest api bitcoin ethereum btc eth neo',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)